# -*- coding: utf-8 -*-

"""
PokeAPI.co API Wrapper
~~~~~~~~~~~~~~~~~~~
A basic wrapper for the PokeAPI.co API.
:copyright: (c) 2019 Lorenzo
:license: MIT, see LICENSE for more details.
"""

__title__ = 'async_pokepy'
__author__ = 'Lorenzo'
__license__ = 'MIT'
__copyright__ = 'Copyright 2019 Lorenzo'
__version__ = '0.0.9a'

from .client import Client
from .exceptions import *
from .types import *



# -*- coding: utf-8 -*-

"""
The MIT License (MIT)

Copyright (c) 2019 Lorenzo

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

import io
from typing import Union

from .http import HTTPPokemonClient
from .types import Pokemon
from .utils import _fmt_param


class Client:
    """The client representing a connection with the API.

    You must use :meth:`~Client.connect` to initiate the class.

    Attributes
    ----------
    loop: :class:`asyncio.AbstractEventLoop`
        The event loop used for HTTP requests.
    """
    __slots__ = ("_http", "loop", "cache_pokemon")

    def __init__(self, http_client: HTTPPokemonClient):
        self._http = http_client
        self.loop = http_client.loop

        self.clear()

    @classmethod
    async def connect(cls, base: str = None, user_agent: str = None, *, loop=None):
        """Connect to the PokeAPI.

        You **must** use this classmethod to connect.

        Parameters
        ----------
        base: Optional[:class:`str`]
            The base to use for all API requests, userful to edit if you
            want to host your own instance of the API.
            Defaults to `https://pokeapi.co/api/v2`.
        user_agent: Optional[:class:`str`]
            The User-Agent header to use when making requests.
        loop: Optional[:class:`asyncio.AbstractEventLoop`]
            The event loop used for HTTP requests, if no loop is provided
            :func:`asyncio.get_event_loop` is used to get one."""
        http = HTTPPokemonClient(loop=loop, user_agent=user_agent, base=base)
        await http.connect()

        return cls(http)

    def clear(self):
        """Clear the cache."""
        self.cache_pokemon = {}

    async def close(self):
        """Close the connection to the API and clear the cache.

        Use this when cleaning up."""
        self.clear()
        await self._http.close()
        del self

    @staticmethod
    def _add_to_cache(cache: dict, obj: Union[Pokemon]):
        cache[(_fmt_param(obj.name), obj.id)] = obj

    @staticmethod
    def _get_from_cache(cache: dict, query: Union[int, str]):
        cache_friendly_name = None

        if isinstance(query, str):
            if query.isdigit():
                cache_friendly_name = int(query)
            else:
                cache_friendly_name = _fmt_param(query)

        for key, value in cache.items():  # Necessary for searching by both ID and name.
            if (cache_friendly_name or query) in key:
                return value

        return None

    async def get_pokemon(self, query: Union[int, str]) -> Pokemon:
        """Get a :class:`Pokemon` from the API.

        The Pokémon will be cached.

        Parameters
        ----------
        query: Union[:class:`int`, :class:`str`]
            The name or id of the Pokèmon.

        Raises
        ------
        PokeAPIException
            The request failed.
        NotFound
            The Pokémon was not found.
        RateLimited
            More then 100 requests in one minute.

        Returns
        -------
        :class:`Pokemon`
            The Pokèmon searched for."""

        check = self._get_from_cache(self.cache_pokemon, query)
        if check:
            return check

        data = await self._http.get_pokemon(query)

        ret = Pokemon(data)
        self._add_to_cache(self.cache_pokemon, ret)

        return ret

    async def save_sprite(self, url: str, fp, *, seek_begin: bool = True) -> int:
        """Save a sprite url into a file-like object.

        Parameters
        ----------
        url: :class:`str`
            The image url of the sprite.
        fp: Union[:class:`io.IOBase`, :class:`os.PathLike`]
            The file-like object to save the sprite to.
            This can be both a path to a file or a BinaryIO.
        seek_begin: :class:`bool`
            Whether to seek to the beginning of the file after saving is done.

        Returns
        -------
        :class:`int`
            The number of bytes written.
        """
        data = await self.read_sprite(url)

        if isinstance(fp, io.IOBase) and fp.writable():
            written = fp.write(data)

            if seek_begin:
                fp.seek(0)

            return written

        with open(fp, "wb") as f:
            return f.write(data)

    async def read_sprite(self, url: str) -> bytes:
        """Read a sprite url's sprite.

        Parameters
        ----------
        url: :class:`str`
            The image url of the sprite.

        Returns
        -------
        :class:`bytes`
            The bytes read.
        """
        return await self._http.download_sprite(url)



# -*- coding: utf-8 -*-

"""
The MIT License (MIT)

Copyright (c) 2019 Lorenzo

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

import asyncio
import logging
import sys
from typing import Coroutine, Union

import aiohttp

from .exceptions import Forbidden, NotFound, PokeAPIException, RateLimited
from .utils import _fmt_param

LOG = logging.getLogger(__name__)


class Route:
    __slots__ = ("params", "url")

    def __init__(self, base, path, *args):
        self.params = [_fmt_param(arg) for arg in args]

        self.url = (base + path + "/".join(self.params))


class HTTPPokemonClient:
    __slots__ = ("loop", "headers", "_session", "_lock", "base")

    def __init__(self, loop: asyncio.AbstractEventLoop = None, base: str = None, user_agent: str = None):
        self.loop = loop or asyncio.get_event_loop()
        self.headers = {
            "User-Agent": user_agent or "Python/{0[0]}.{0[1]} aiohttp/{1}".format(sys.version_info, aiohttp.__version__)
        }
        self._session = None
        self._lock = asyncio.Lock(loop=self.loop)
        self.base = base or "https://pokeapi.co/api/v2"

    async def request(self, route, **params) -> Union[str, dict]:
        async with self._lock:
            for tries in range(5):
                async with self._session.get(route.url, **params) as resp:
                    LOG.info("%s %s returned %d status code", resp.method, resp.url, resp.status)

                    data = (await resp.json() if "application/json" in resp.headers["Content-Type"]
                            else await resp.text())

                    if 300 > resp.status >= 200:
                        LOG.debug("%s %s succeeded with data %s", resp.method, resp.url, data)

                        return data

                    if resp.status == 429:
                        LOG.error("Surpassed 100 API requests in one minute")

                        raise RateLimited(resp, "Surpassed 100 API requests in one minute.")

                    if resp.status in {500, 502}:
                        sleep_time = 1 + tries * 2
                        LOG.warning("Internal API error, retrying in %d", sleep_time)

                        await asyncio.sleep(sleep_time, loop=self.loop)
                        continue

                    if resp.status == 403:
                        raise Forbidden(resp, "Forbidden endpoint.")
                    if resp.status == 404:
                        raise NotFound(resp, "Endpoint not found.")

                    raise PokeAPIException(resp, "Uncaught status code.")

            LOG.critical("Request timed out")
            raise PokeAPIException(resp, "Request timed out.")

    async def connect(self):
        self._session = aiohttp.ClientSession(headers=self.headers, loop=self.loop)

    async def close(self):
        await self._session.close()
        del self

    async def download_sprite(self, url: str) -> bytes:
        async with self._session.get(url) as resp:
            if resp.status == 200:
                return await resp.read()
            if resp.status == 404:
                raise NotFound(resp, "Sprite not found.")
            if resp.status == 403:
                raise Forbidden(resp, "Cannot retrieve sprite.")

            raise PokeAPIException(resp, "Failed to get sprite.")

    def get_pokemon(self, query: Union[int, str]) -> Coroutine:
        return self.request(Route(self.base, "/pokemon/", query))

    def get_move(self, query: Union[int, str]) -> Coroutine:
        return self.request(Route(self.base, "/move/", query))

    def get_ability(self, query: Union[int, str]) -> Coroutine:
        return self.request(Route(self.base, "/ability/", query))



import re

from setuptools import setup

with open("README.md") as f:
    long_description = f.read()

with open("requirements.txt") as f:
    requirements = f.read().splitlines()

with open("async_pokepy/__init__.py") as f:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', f.read(), re.MULTILINE).group(1)

if not version:
    raise RuntimeError('version is not set')

extra = {
    "docs": [
        "sphinx",
        "sphinxcontrib.napoleon",
        "sphinxcontrib-asyncio",
        "pallets_sphinx_themes"
    ],
}

setup(name="async_pokepy",
      version=version,
      description="A simple asynchronous wrapper for the PokeAPI.co API.",
      long_description=long_description,
      long_description_content_type="text/markdown",
      url="https://github.com/PendragonLore/async_pokepy",
      python_requires=">=3.5.3",
      install_requires=requirements,
      include_package_data=True,
      license="MIT",
      extras_require=extra,
      packages=["async_pokepy", "async_pokepy.types"],
      classifiers=[
          "Programming Language :: Python :: 3.6",
          "Programming Language :: Python :: 3.7",
          "License :: OSI Approved :: MIT License",
          "Natural Language :: English",
          "Topic :: Internet",
          "Topic :: Software Development :: Libraries",
          "Topic :: Software Development :: Libraries :: Python Modules",
          "Topic :: Utilities",
          "Operating System :: OS Independent",
          "Development Status :: 3 - Alpha",
      ],
      )



# -*- coding: utf-8 -*-

"""
The MIT License (MIT)

Copyright (c) 2019 Lorenzo

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from ..utils import _pretty_format
from ._base import BaseObject
from .other import VersionGameIndex


class Pokemon(BaseObject):
    """Represents a Pokémon object.

    .. container:: operations

        .. describe:: str(x)

            Returns the Pokémon's name.

        .. describe:: x[y]

            Returns a Pokémon's y attribute.

        .. describe:: x == y

            Check if two Pokémons are the same.

        .. describe:: x != y

            Check if two Pokémons are *not* the same.

    Attributes
    ----------
    id: :class:`int`
        The PokeAPI identifier for this Pokémon.
    name: :class:`str`
        The name for this Pokémon.
    base_experience: :class:`int`
        The base experience gained for defeating this Pokémon.
    height: :class:`int`
        The height of this Pokémon, in decimetres.
    is_default: :class:`bool`
        ``True`` if this Pokémon is used as the default for each species.
    order: :class:`int`
        Order for sorting. Almost national order, except families are grouped together.
    weight: :class:`id`
        The weight of this Pokémon, in hectograms.
    abilities: List[:class:`PokemonAbility`]
        A list of :class:`PokemonAbility` this Pokémon could potentially have.
    moves: List[:class:`PokemonMove`]
        A list of moves along with learn methods and level details pertaining to specific version groups.
    forms: List[:class:`str`]
        A list of forms this Pokémon can take on.
    game_indices: List[:class:`VersionGameIndex`]
        A list of :class:`VersionGameIndex` relevent to Pokémon item by generation.
    stats: List[:class:`PokemonStat`]
        A list of base stat values for this Pokémon.
    types:  List[:class:`PokemonType`]
        A list of details showing types this Pokémon has.
    sprites: :class:`PokemonSprites`
        A set of sprites used to depict this Pokémon in the game.
    """
    __slots__ = (
        "stats", "types", "weight", "moves", "abilities", "height",
        "sprites", "held_items", "base_experience", "is_default", "order",
        "species", "forms", "game_indices"
    )

    def __init__(self, data: dict):
        super().__init__(data)

        self.base_experience = data["base_experience"]
        self.weight = data["weight"]
        self.height = data["height"]
        self.stats = data["stats"]
        self.is_default = data["is_default"]
        self.order = data["order"]

        self.species = data["species"]["name"]
        self.forms = [_pretty_format(d["name"]) for d in data["forms"]]

        self.sprites = PokemonSprites(data["sprites"])

        self.abilities = [PokemonAbility(d) for d in data["abilities"]]
        self.types = [PokemonType(d) for d in data["types"]]
        self.moves = [PokemonMove(d) for d in data["moves"]]
        self.stats = [PokemonStat(d) for d in data["stats"]]
        self.held_items = [PokemonHeldItem(d) for d in data["held_items"]]
        self.game_indices = [VersionGameIndex(d) for d in data["game_indices"]]

    def __repr__(self) -> str:
        return "<Pokemon id={0.id} name='{0}'>".format(self)


class PokemonStat:
    """Represents a partial Stat object bound to a :class:`Pokemon`.

    Attributes
    ----------
    stat: :class:`str`
        The name of the stat.
    effort: :class:`int`
        The effort points (EV) the Pokémon has in the stat.
    base_stat: :class:`int`
        The base value of the stat.
    """
    __slots__ = ("stat", "effort", "base_stat")

    def __init__(self, data: dict):
        self.stat = _pretty_format(data["stat"]["name"])
        self.effort = data["effort"]
        self.base_stat = data["base_stat"]

    def __repr__(self):
        return "<PokemonStat stat='{0.stat}' effort={0.effort} base_stat={0.base_stat}>".format(self)


class PokemonAbility:
    """Represents a partial Ability object bound to a :class:`Pokemon`.

    .. container:: operations

        .. describe:: str(x)

            Returns the ability's name.

    Attributes
    ----------
    ability: :class:`str`
        The ability's name.
    slot: :class:`int`
        The slot this ability occupies in the Pokémon species.
    is_hidden: :class:`bool`
        Whether or not this is a hidden ability."""
    __slots__ = ("is_hidden", "slot", "ability")

    def __init__(self, data: dict):
        self.is_hidden = data["is_hidden"]
        self.slot = data["slot"]
        self.ability = _pretty_format(data["ability"]["name"])

    def __str__(self) -> str:
        return self.ability

    def __repr__(self) -> str:
        return "<PokemonAbility ability='{0}' is_hidden={0.is_hidden} slot={0.slot}>".format(self)


class PokemonType:
    """Represents a partial Type object bound to a :class:`Pokemon`.

    .. container:: operations

        .. describe:: str(x)

            Returns the type's name.

    Attributes
    ----------
    type: :class:`str`
        The name of the type the Pokémon has.
    slot: :class:`int`
        The order the Pokémon's types are listed in."""
    __slots__ = ("type", "slot")

    def __init__(self, data: dict):
        self.type = _pretty_format(data["type"]["name"])
        self.slot = data["slot"]

    def __str__(self) -> str:
        return self.type

    def __repr__(self):
        return "<PokemonType type='{0}' slot={0.slot}>".format(self)


class PokemonSprites:
    """Represents all of the possible sprites a :class:`Pokemon` could have.

    .. note::
        All of these attributes could be ``None`` or a friendly image url of the described sprite.

    Attributes
    ----------
    front_default: Optional[:class:`str`]
        The default sprite of a Pokémon from the front in battle.
    front_shiny: Optional[:class:`str`]
        The shiny sprite of a Pokémon from the front in battle.
    front_female: Optional[:class:`str`]
        The female sprite of a Pokémon from the front in battle.
    front_shiny_female: Optional[:class:`str`]
        The shiny female sprite of a Pokémon from the front in battle.
    back_default: Optional[:class:`str`]
        The default sprite of a Pokémon from the back in battle.
    back_shiny: Optional[:class:`str`]
        The shiny sprite of a Pokémon from the back in battle.
    back_female: Optional[:class:`str`]
        The female sprite of a Pokémon from the back in battle.
    back_shiny_female: Optional[:class:`str`]
        The shiny sprite depiction of a Pokémon from the back in battle.
    """
    __slots__ = (
        "front_default", "front_shiny", "front_female", "front_shiny_female",
        "back_default", "back_shiny", "back_female", "back_shiny_female"
    )

    def __init__(self, data: dict):
        self.front_default = data["front_default"]
        self.front_shiny = data["front_shiny"]
        self.front_female = data["front_female"]
        self.front_shiny_female = data["front_shiny_female"]

        self.back_default = data["back_default"]
        self.back_shiny = data["back_shiny"]
        self.back_female = data["back_female"]
        self.back_shiny_female = data["back_shiny_female"]


class PokemonMove:
    """Represents a partial Move object bound to a :class:`Pokemon`

    .. container:: operations

        .. describe:: str(x)

            Returns the move's name.

    Attributes
    ----------
    move: :class:`str`
        The name of the move the Pokémon can learn.
    version_group_details: List[:class:`PokemonMoveVersion`]
        The details of the version in which the Pokémon can learn the move.
    """
    __slots__ = ("move", "version_group_details")

    def __init__(self, data: dict):
        self.move = _pretty_format(data["move"]["name"])
        self.version_group_details = [PokemonMoveVersion(d) for d in data["version_group_details"]]

    def __str__(self) -> str:
        return self.move

    def __repr__(self) -> str:
        return "<PokemonMove move='{0.move}' version_group_details={0.version_group_details}>".format(self)


class PokemonMoveVersion:
    """Represents a partial Version object bound to a :class:`PokemonMove`.

    Attributes
    ----------
    move_learn_method: :class:`str`
        The name of the method by which the move is learned.
    version_group: :class:`str`
        The name of the version group in which the move is learned.
    level_learned_at: :class:`int`
        The minimum level to learn the move."""
    __slots__ = ("move_learn_method", "version_group", "level_learned_at")

    def __init__(self, data: dict):
        self.level_learned_at = data["level_learned_at"]
        self.version_group = _pretty_format(data["version_group"]["name"])
        self.move_learn_method = _pretty_format(data["move_learn_method"]["name"])

    def __repr__(self) -> str:
        return ("<PokemonMoveVersion move_learn_method='{0.move_learn_method}' level_learned_at={0.level_learned_at}>"
                .format(self))


class PokemonHeldItem:
    """Represents a partial Item object bound to a :class:`Pokemon`.

    .. container:: operations

        .. describe:: str(x)

            Returns the ability's name.

    Attributes
    ----------
    item: :class:`str`
        The item the bound Pokémon holds.
    version_details: List[:class:`PokemonHeldItemVersion`]
        The details of the different versions in which the item is held."""
    __slots__ = ("item", "version_details")

    def __init__(self, data: dict):
        self.item = _pretty_format(data["item"]["name"])
        self.version_details = [PokemonHeldItemVersion(d) for d in data["version_details"]]

    def __str__(self) -> str:
        return self.item

    def __repr__(self) -> str:
        return "<PokemonHeldItem item='{0}' version_details={0.version_details}>".format(self)


class PokemonHeldItemVersion:
    """Represents a partial Version object bound to a :class:`PokemonHeldItem`.

    Attributes
    ----------
    version: :class:`str`
        The version bound to the held item.
    rarity: :class:`int`
        How often the item is held."""
    __slots__ = ("version", "rarity")

    def __init__(self, data: dict):
        self.version = _pretty_format(data["version"]["name"])
        self.rarity = data["rarity"]

    def __repr__(self) -> str:
        return "<PokemonHeldItemVersion version={0.version} rarity={0.rarity}>".format(self)



# -*- coding: utf-8 -*-

"""
The MIT License (MIT)

Copyright (c) 2019 Lorenzo

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""


class PokemonException(Exception):
    """The base exception for this wrapper."""


class PokeAPIException(PokemonException):
    """Exception raised when an HTTP request is not successful.

    Attributes
    ----------
    response: :class:`aiohttp.ClientResponse`
        The failed HTTP response.
    status: :class:`int`
        The HTTP status code.
    message: :class:`str`
        A, hopefully, useful exception message."""

    def __init__(self, response, message: str):
        self.response = response
        self.status = response.status

        self.message = "API responded with status code {0} {2}: {1}".format(self.status, message, response.reason)

        super().__init__(self.message)


class RateLimited(PokeAPIException):
    """Exception raised when an HTTP request is equal to 429 TOO MANY REQUESTS.

    This exception will only raise if you surpass 100 requests per minutes,
    if you need more requests then that it would be better to host your own
    API instance.

    This inherits from :exc:`PokeAPIException`."""


class NotFound(PokeAPIException):
    """Exception raised when an HTTP request response code is equal to 404 NOT FOUND.

    This inherits from :exc:`PokeAPIException`."""


class Forbidden(PokeAPIException):
    """Exception raised when an HTTP request response code is equal to 403 FORBIDDEN.

    This inherits from :exc:`PokeAPIException`."""
